import abc
import six
from collections import namedtuple
import numpy as np
from toon.input.clock import mono_clock
import inspect
import ctypes


@six.add_metaclass(abc.ABCMeta)
class Obs():
    """Abstract base class for observations.

    This is subclassed when making new subclasses of toon.input.BaseDevice,
    and is used to preallocate shared memory between the device and main processes.
    """
    @property
    @abc.abstractmethod
    def shape(self):
        """Shape of the observation."""
        return None

    @property
    @abc.abstractmethod
    def ctype(self):
        """Data type of the observation. Can be built-in type (e.g. int, float),
        numpy types, or ctypes types.
        """
        return None

    def __init__(self, time, data):
        """Create a new Observation.

        Parameters
        ----------
        time: float
            Time that the data was observed.
        data: array_like
            Observed data. Must match the shape of the subclass.
        """
        try:
            self.time = float(time)  # what if time is not a double?
            # is reshape expensive? should we just trust they did it right?
            self.data = np.asarray(data, dtype=self.ctype)
            self.data.shape = self.shape  # will error if mismatch?
        except (TypeError, ValueError):  # time or data not right. TODO: should we error out here instead?
            self.time = None
            self.data = None

    def __repr__(self):
        return 'type: %s\ntime: %f\ndata: %s\nshape: %s\nctype: %s' % (type(self).__name__, self.time, self.data, self.shape, self.ctype)

    def __str__(self):
        return '%s(time: %f, data: %s)' % (type(self).__name__, self.time, self.data)

    def any(self):
        """Helper to check whether there is any data."""
        return not (self.time is None or self.data is None)


@six.add_metaclass(abc.ABCMeta)
class BaseDevice():
    """Abstract base class for input devices.

    Attributes
    ----------
    sampling_frequency: int
        Expected sampling frequency of the device, used by toon.input.MpDevice for preallocation.
        We preallocate for 1 second of data (e.g. 500 samples for a sampling_frequency of 500 Hz).
    """

    def __init__(self, clock=mono_clock.get_time):
        """Create new device.

        Parameters
        ----------
        clock: function or method
            The clock used for timestamping events. Defaults to toon.input.mono_clock, which
            allows us to share a reference time between the parent and child processes. The 
            mono_clock is based off psychopy.clock.MonotonicClock 
            (on Windows, time.perf_counter seems to be relative to when the process is created, 
            which makes it difficult to relate the time between processes).
        """
        _obs = self.__class__.get_obs()
        self.Returns = BaseDevice.build_named_tuple(_obs)
        self.clock = clock

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def read(self):
        pass

    sampling_frequency = 500

    # helpers to figure out the data returned by device
    # (without instantiation--key b/c we need to do *before* we instantiate on other process)
    @classmethod
    def get_obs(cls):
        # get all user-defined Obs defined within the class (as long as they don't start w/ double underscore)
        # separated from tuple building so that
        return [getattr(cls, p) for p in dir(cls) if not p.startswith('__')
                and not p.startswith('_abc')
                and not isinstance(getattr(cls, p), property)
                and inspect.isclass(getattr(cls, p))
                and issubclass(getattr(cls, p), Obs)]

    @classmethod
    def build_named_tuple(cls, obs):
        if obs:
            class Returns(namedtuple('Returns', [x.__name__.lower() for x in obs])):
                def any(self):
                    # simplify user checking of whether there's any data
                    return any([x is not None and x.time is not None for x in self])
            # default values of namedtuple to None (see mouse.py for example why)
            Returns.__new__.__defaults__ = (None,) * len(Returns._fields)
            return Returns
        return None



import ctypes
import gc
import multiprocessing as mp
import os
from collections import namedtuple
from copy import copy
from itertools import compress
from sys import platform
from toon.input.tsarray import TsArray as TsA

import numpy as np
import psutil

ctypes_map = np.ctypeslib._typecodes.copy()
ctypes_map['|S1'] = ctypes.c_char  # TODO: I have no clue if this is valid
ctypes_map['|b1'] = ctypes.c_bool


def shared_to_numpy(mp_arr, dims, dtype):
    """Convert a :class:`multiprocessing.Array` to a numpy array.
    Helper function to allow use of a :class:`multiprocessing.Array` as a numpy array.
    Derived from the answer at:
    <https://stackoverflow.com/questions/7894791/use-numpy-array-in-shared-memory-for-multiprocessing>
    """
    return np.frombuffer(mp_arr, dtype=dtype).reshape(dims)


class MpDevice(object):
    """Creates and manages a process for polling an input device."""

    def __init__(self, device, high_priority=True,
                 buffer_len=None, **device_kwargs):
        """Create a new MpDevice.

        Parameters
        ----------
        device: class (derived from toon.input.BaseDevice)
            Class of input device. We will instantiate the device on the child process.
        high_priority: bool
            Whether the priority of the child process should be elevated.
        buffer_len: int, optional
            Overrides the device's sampling_frequency when specifing the size of the
            circular buffer.
        **device_kwargs
            Keyword arguments, passed to the device during instantiation.
        """
        self.device = device
        self.device_kwargs = device_kwargs
        self.buffer_len = buffer_len
        self.high_priority = high_priority

    def start(self):
        """Start polling from the device on the child process.

        Allocates all resources and creates the child process.

        Notes
        -----
        Prefer using as a context manager over explicitly starting and stopping.

        Raises
        ------
        Will raise an exception if something goes wrong during instantiation
        of the device.
        """
        # For Macs, use spawn (interaction with OpenGL or ??)
        # Windows only does spawn
        if platform == 'darwin' or platform == 'win32':
            try:
                mp.set_start_method('spawn')
            except (AttributeError, RuntimeError):
                pass  # already started a process, or on python2

        n_buffers = 2
        self.shared_locks = []
        # make one lock per buffer
        for i in range(n_buffers):
            self.shared_locks.append(mp.Lock())
        self.remote_ready = mp.Event()  # signal to main process that remote is done setup
        self.kill_remote = mp.Event()  # signal to remote process to die
        self.remote_done = mp.Event()
        self.local_err, remote_err = mp.Pipe(duplex=False)
        self.current_buffer_index = mp.Value(ctypes.c_bool, 0, lock=False)  # shouldn't need a lock

        # figure out number of observations to save between reads
        nrow = 100
        if self.device.sampling_frequency:
            nrow = self.device.sampling_frequency
        # if the user specified a runtime sampling_frequency, use that
        kwarg_freq = self.device_kwargs.get('sampling_frequency', None)
        if kwarg_freq:
            nrow = kwarg_freq
        if self.buffer_len:  # buffer_len overcomes all
            nrow = self.buffer_len
        nrow = max(int(nrow), 1)  # make sure we have at least one row
        _device_obs = self.device.get_obs()
        # use this tuple to return data
        self._return_tuple = self.device.build_named_tuple(_device_obs)
        self.nt = namedtuple('obs', 'time data')
        self._data = [[] for i in range(n_buffers)]  # one set of data per buffer
        # for each observation type, allocate arrays (though note that
        # there's nothing useful in them right now)
        for i in range(n_buffers):
            for obs in _device_obs:
                self._data[i].append(DataGlob(obs.ctype, obs.shape, nrow, self.shared_locks[i]))

        self._res = [None] * len(self._data[0])

        self.process = mp.Process(target=remote,
                                  args=(self.device, self.device_kwargs, self._data,
                                        self.remote_ready, self.kill_remote, os.getpid(),
                                        self.current_buffer_index, self.remote_done, remote_err))
        self.process.daemon = True
        self.process.start()
        self.ps_process = psutil.Process(self.process.pid)
        self.original_nice = self.ps_process.nice()
        self.set_high_priority(self.high_priority)  # lame try to set priority
        for i in range(n_buffers):
            for obs in self._data[i]:
                obs.generate_squeeze()
        self.check_error()
        self.remote_ready.wait()  # pause until the remote says it's ready

    def read(self):
        """Retrieve all observations that have occurred since the last read.

        Notes
        ----
        The data is stored in a circular buffer, which means that if the number
        of observations since the last read exceeds the preallocated data, the oldest
        data will be overwritten in favor of the newest. If this behavior is undesirable,
        either bump the sampling_frequency of the Device or the buffer_len of the MpDevice
        up to an adequate number, depending on the sampling rate of the device and how
        often you expect to call read().

        We minimize copies by default (i.e. we return a view of the local data).
        This means that in certain situations, such as appending the data to a list,
        there may be unexpected results (as all elements would be a reference to the
        same object/array). If you run into this, try explicitly `copy()`ing the data.

        If a named tuple is returned, the names are in *alphabetical* order, not the order
        that they are specified in the Device. This may come up if unpacking the read.

        Returns
        -------
        toon.input.TsArray (i.e. a numpy array with a `time` attribute) for a single Observation, or
        named tuple of TsArrays (if multiple types of Observations), where the names match the subclassed
        `Obs`ervations from the Device.

        Raises
        ------
        May raise an exception if one has occurred on the child process since the last read.
        """

        # get the currently used buffer
        # note that the value only changes *if* this function has acquired the
        # lock, so we should always be safe to access w/o lock here

        # check if error, and raise if present
        self.check_error()

        current_buffer_index = int(self.current_buffer_index.value)
        # this *may* block, if the remote is currently writing
        with self._data[current_buffer_index][0].lock:
            for counter, datum in enumerate(self._data[current_buffer_index]):
                datum.local_count = datum.counter.value
                datum.counter.value = 0  # reset (so that we start writing to top of array)
                if datum.local_count > 0:
                    np.copyto(datum.local_data.time, datum.np_data.time)
                    np.copyto(datum.local_data.T, datum.np_data.data.T)  # allow copying to 1D
                    self._res[counter] = datum.local_data[0:datum.local_count]
                else:
                    self._res[counter] = None
        # if only one Obs, don't stuff into a namedtuple
        if len(self._res) == 1:
            return self._res[0]
        # plug values into namedtuple
        return self._return_tuple(*self._res)

    def stop(self):
        """Stop reading from the device and kill the child process.

        Notes
        -----
        Prefer using as a context manager over explicitly starting and stopping.
        """
        self.set_high_priority(False)
        self.kill_remote.set()
        self.remote_done.wait()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()

    def check_error(self):
        """See if any exceptions have occurred on the child process, or whether
        the device was already closed.
        """
        if self.remote_done.is_set():
            # already closed (e.g. if using start() & stop() manually)
            if self.kill_remote.is_set():
                raise ValueError('MpDevice is closed.')
            # real error at *any* point remotely (device instantiation, shape mismatch, ...)
            raise self.local_err.recv()

    def set_high_priority(self, val=False):
        """Set the priority of the child process.
        Fails unless the user has adquate permissions.
        """
        try:
            if val:
                if psutil.WINDOWS:
                    self.ps_process.nice(psutil.HIGH_PRIORITY_CLASS)
                else:
                    self.ps_process.nice(-10)
            else:
                self.ps_process.nice(self.original_nice)
        except (psutil.AccessDenied, psutil.NoSuchProcess) as e:
            pass


def remote(device, device_kwargs, shared_data,
           # extras
           remote_ready, kill_remote, parent_pid,  # don't include in coverage, implicitly tested
           current_buffer_index, remote_done, remote_err):  # pragma: no cover
    """Poll the device for data."""

    def process_data(shared, datum):
        if shared.counter.value < shared.nrow:  # haven't filled buffer yet, so next available row
            next_index = shared.counter.value
            shared.np_data.time[next_index] = datum.time
            shared.np_data.data[next_index, :] = datum.data
        else:  # rolling buffer, had cursed my bedroom
            np.copyto(shared.np_data.time, np.roll(shared.np_data.time, -1, axis=0))
            shared.np_data.time[-1] = datum.time
            np.copyto(shared.np_data.data, np.roll(shared.np_data.data, -1, axis=0))
            shared.np_data.data[-1, :] = datum.data
        # successful read, increment the indexing counter for this stream of data
        shared.counter.value += 1

    try:
        gc.disable()
        # instantiate the device
        dev = device(**device_kwargs)

        for i in shared_data:
            for j in i:
                j.generate_np_version()

        with dev:
            remote_ready.set()  # signal to the local process that remote is ready to go
            while not kill_remote.is_set() and psutil.pid_exists(parent_pid):
                data = dev.read()  # get observation(s) from device
                buffer_index = int(current_buffer_index.value)  # can only change later
                if isinstance(data, list):  # if a list of observations, rather than a single one
                    inds = [[d is not None and d.any() for d in l] for l in data]
                    flag = any([any(d) for d in inds])
                    is_list = True
                else:
                    inds = [d is not None and d.any() for d in data]
                    flag = any(inds)
                    is_list = False
                if flag:  # any data at all (otherwise, don't bother acquiring locks)
                    # test whether the current buffer is accessible
                    lck = shared_data[buffer_index][0].lock
                    success = lck.acquire(block=False)
                    if not success:  # switch to the other buffer (the local process is in the midst of reading)
                        current_buffer_index.value = not current_buffer_index.value
                        buffer_index = int(current_buffer_index.value)
                        lck = shared_data[buffer_index][0].lock
                        lck.acquire()
                    try:
                        if is_list:
                            for dat, ind in zip(data, inds):
                                for counter, datum in zip(np.where(ind)[0], compress(dat, ind)):
                                    process_data(shared_data[buffer_index][counter], datum)
                        else:
                            for counter, datum in zip(np.where(inds)[0], compress(data, inds)):
                                process_data(shared_data[buffer_index][counter], datum)
                    finally:
                        lck.release()
    except Exception as e:
        remote_err.send(e)
        gc.enable()
        remote_err.close()
    finally:
        remote_ready.set()
        remote_done.set()


# make sure this is visible for pickleability
obs = namedtuple('obs', 'time data')


class DataGlob(object):
    def __init__(self, ctype, shape, nrow, lock):
        self.new_dims = (nrow,) + shape
        self.ctype = ctypes_map[np.dtype(ctype).str]
        self.shape = shape
        self.is_scalar = self.shape == (1,)
        self.lock = lock
        prod = int(np.prod(self.new_dims))
        # don't touch (usually)
        self.nrow = int(nrow)
        self._mp_data = obs(time=mp.Array(ctypes.c_double, self.nrow, lock=False),
                            data=mp.Array(self.ctype, prod, lock=False))
        self.counter = mp.Value(ctypes.c_uint, 0, lock=False)
        self.generate_np_version()
        self.generate_local_version()

    def generate_np_version(self):
        self.np_data = obs(time=shared_to_numpy(self._mp_data.time, (self.nrow,), ctypes.c_double),
                           data=shared_to_numpy(self._mp_data.data, self.new_dims, self.ctype))

    def generate_local_version(self):
        self.local_count = 0
        self.local_data = TsA(self.np_data.data.copy(), time=self.np_data.time.copy())

    def generate_squeeze(self):
        # if scalar data, give the user a 1D array (rather than 2D)
        if self.is_scalar:
            self.local_data = TsA(np.squeeze(self.np_data.data.copy()), time=self.np_data.time.copy())



import numpy as np
from toon.input.tsarray import TsArray


def test_creation():
    arr = TsArray([1, 2, 3], time=[1, 2, 3])
    # can make it
    assert(len(arr) == len(arr.time))
    nparr = np.array([1, 2, 3])
    # stores vals like numpy
    assert(all(nparr == arr))


def test_slice():
    arr = TsArray([[1, 2, 3], [4, 5, 6],
                   [7, 8, 9], [10, 11, 12]],
                  time=[0.1, 0.2, 0.3, 0.4])

    assert(all(arr[-1] == [10, 11, 12]))
    assert(arr[-1].time == 0.4)
    # non-scalar
    assert(all(arr[-2:].time == [0.3, 0.4]))
    # multidimensional slice (should take the first axis)
    assert(all(arr[-2:, :].time == [0.3, 0.4]))
    assert(all(arr[-2:].time == arr.time[-2:]))


def test_copy():
    arr = TsArray([[1, 2, 3], [4, 5, 6],
                   [7, 8, 9], [10, 11, 12]],
                  time=[0.1, 0.2, 0.3, 0.4])
    arr2 = arr.copy()
    assert(all(arr2.time == [0.1, 0.2, 0.3, 0.4]))
    # plug in an index (used to carry over to new objs)
    arr[1]
    arr2 = arr.copy()
    assert(all(arr2.time == [0.1, 0.2, 0.3, 0.4]))



import struct
import serial
import numpy as np
from time import sleep
from toon.input.device import BaseDevice, Obs

# thanks to ROS http://docs.ros.org/fuerte/api/cyberglove/html/serial__glove_8hpp.html
# for commands


class Cyberglove(BaseDevice):
    sampling_frequency = 150

    class Pos(Obs):
        # TODO: split into fingers/joints?
        shape = (18,)
        ctype = float

    def __init__(self, port, **kwargs):
        super(Cyberglove, self).__init__(**kwargs)
        self.port = port  # TODO: auto-detect using serial.tools.list_ports
        self.dev = None

    def __enter__(self):
        self.dev = serial.Serial(self.port, 115200, timeout=0.02)
        self.dev.reset_input_buffer()
        sleep(0.1)
        # test whether the device is connected/on
        self.dev.write(b'l ?\r')
        # should echo back & give result
        if not self.dev.read():
            raise ValueError('Make sure the device is switched on.')
        self.dev.write(b'f 0\r')  # stop filtering
        self.dev.write(b't 1152 1\r')  # 100 Hz
        self.dev.write(b'u 0\r')  # don't transmit status
        self.dev.write(b'l 1\r')  # light on
        sleep(0.1)
        self.dev.reset_input_buffer()
        self.dev.write(b'S')  # start streaming
        self.dev.flush()
        # not sure how to get rid of the rest of the garbage,
        # so spin until we find the first b'S'
        for i in range(40):
            tmp = self.dev.read()
            if tmp == b'S':
                self.dev.read(19)  # read the rest of the line
                break
        if i >= 39:
            raise ValueError('Did not find the start byte.')
        return self

    def read(self):
        # one byte (S)
        val = self.dev.read(1)
        time = self.clock()
        if val:
            data = self.dev.read(19)  # read remaining bytes ('<data> \x00')
            data = struct.unpack('<' + 'B' * 18, data[:-1])
            data = [(d - 1.0)/254.0 for d in data]
            return self.Returns(self.Pos(time, data))

    def __exit__(self, *args):
        self.dev.write(b'\x03')  # stop streaming
        self.dev.write(b'l 0\r')  # light off
        sleep(0.1)
        self.dev.reset_input_buffer()
        self.dev.reset_output_buffer()
        self.dev.close()



from tests.input.mockdevices import Dummy, DummyList
from toon.input.device import Obs


def test_device_single():
    dev = Dummy()
    res = dev.read()
    assert(issubclass(type(res[0]), Obs))
    assert(issubclass(type(res.num1.time), float))
    assert(len(res.num1.data) == 5)


def test_device_multi():
    dev = DummyList()
    res = dev.read()
    assert(len(res) == 2)


def test_context():
    dev = Dummy()
    with dev:
        res = dev.read()
    assert(issubclass(type(res[0]), Obs))
    assert(issubclass(type(res.num1.time), float))
    assert(len(res.num1.data) == 5)



import numpy as np
from ctypes import c_double
# tons of thanks to
# https://stackoverflow.com/questions/31282764/when-subclassing-a-numpy-ndarray-how-can-i-modify-getitem-properly
# and https://docs.scipy.org/doc/numpy/user/basics.subclassing.html


class TsArray(np.ndarray):
    def __new__(cls, data, time=None):
        obj = np.asarray(data).view(cls)
        obj.time = np.asarray(time, dtype=c_double)
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.time = getattr(obj, 'time', None)
        try:
            self.time = self.time[obj._new_time_index]
        except:
            pass

    def copy(self, **kwargs):
        self._new_time_index = slice(None, None, None)
        return super(TsArray, self).copy(**kwargs)

    def __getitem__(self, item):
        try:
            if isinstance(item, (slice, int)):
                self._new_time_index = item
            else:
                self._new_time_index = item[0]
        except:
            pass
        return super(TsArray, self).__getitem__(item)



from collections import namedtuple
from toon.anim.easing import linear
from toon.anim.interpolators import lerp, select


class Track(object):
    """Storage for (keyframe, value) pairs.

    Also handles interpolation and easing between keyframes.
    """

    def __init__(self, data, interpolator=lerp, easing=linear):
        """Creates a Track.

        Parameters
        ----------
        data: list of (keyframe, value) pairs
        interpolator: function
            `lerp` for linear interpolation, `select` for stepwise behavior.
        easing: function
            Rate of change of the value over time. See toon.anim.easing, or
            specificy a custom function that takes a single parameter, and
            returns a value on the interval [0, 1].
        """
        # data is list of tuples
        self.data = data
        self.interpolator = interpolator
        self.easing = easing
        # if data is non-numeric, force user to use select
        if not isinstance(data[0][1], (float, int)):
            self.interpolator = select
            self.easing = linear

    def at(self, time):
        """Get the interpolated value of a track at a given time.

        Parameters
        ----------
        time: float
            Time of interest.

        Returns
        -------
        The interpolated value at the given point in time.
        """
        # find the two keyframes to interpolate between
        # rel_time is the time relative to the start
        if time < self.data[0][0]:
            # TODO: extrapolation (currently equivalent to constant)
            return self.data[0][1]
        try:
            index, goal = next((i, x) for i, x in enumerate(self.data) if x[0] > time)
        except StopIteration:
            # TODO: Extrapolation (currently equivalent to constant)
            return self.data[-1][1]
        # prev keyframe to check against
        reference = self.data[index - 1]
        # calculate time stuff
        goal_time = goal[0] - reference[0]
        new_time = time - reference[0]
        time_warp = self.easing(1 - ((goal_time - new_time)/goal_time))
        return self.interpolator(reference[1], goal[1], time_warp)

    def duration(self):
        """The maximum duration of the track."""
        # last time in keyframes
        return self.data[-1][0]



from toon.input.device import BaseDevice, Obs
from timeit import default_timer
import ctypes
import numpy as np


class Dummy(BaseDevice):
    counter = 0
    t0 = default_timer()

    class Num1(Obs):
        shape = (5,)
        ctype = ctypes.c_float

    class Num2(Obs):
        shape = (3, 3)
        ctype = ctypes.c_int

    @property
    def foo(self):
        return 3

    def read(self):
        dat = None
        self.counter += 1
        if self.counter % 10 == 0:
            dat = np.random.randint(5, size=self.Num2.shape)
        while default_timer() - self.t0 < (1.0/self.sampling_frequency):
            pass
        self.t0 = default_timer()
        t = self.clock()
        return self.Returns(num1=self.Num1(t, np.random.random(self.Num1.shape)),
                            num2=self.Num2(t, dat))


class SingleResp(BaseDevice):
    t0 = default_timer()
    sampling_frequency = 1000
    counter = 0

    class Num1(Obs):
        shape = (1,)
        ctype = int

    def read(self):
        while default_timer() - self.t0 < (1.0/self.sampling_frequency):
            pass
        val = self.counter
        self.counter += 1
        self.t0 = default_timer()
        t = self.clock()
        return self.Returns(self.Num1(t, val))


class Timebomb(SingleResp):
    def read(self):
        while default_timer() - self.t0 < (1.0/self.sampling_frequency):
            pass
        val = self.counter
        self.counter += 1
        if self.counter > 10:
            raise ValueError('Broke it.')
        self.t0 = default_timer()
        t = self.clock()
        return self.Returns(self.Num1(t, val))


class DummyList(BaseDevice):
    counter = 0
    t0 = default_timer()

    class Num1(Obs):
        shape = (5,)
        ctype = ctypes.c_float

    class Num2(Obs):
        shape = (3, 3)
        ctype = ctypes.c_int

    def read(self):
        dat = None
        self.counter += 1
        if self.counter % 10 == 0:
            dat = np.random.randint(5, size=self.Num2.shape)
        while default_timer() - self.t0 < (1.0/self.sampling_frequency):
            pass
        self.t0 = default_timer()
        t = self.clock()
        return [self.Returns(num1=self.Num1(t, np.random.random(self.Num1.shape)),
                             num2=self.Num2(t, dat)),
                self.Returns(num1=self.Num1(t, np.random.random(self.Num1.shape)),
                             num2=self.Num2(t, dat))]


if __name__ == '__main__':  # pragma: no cover
    from time import time, sleep
    from timeit import default_timer
    import matplotlib.pyplot as plt
    from toon.input.mpdevice import MpDevice
    Dummy.sampling_frequency = 1000
    dev = MpDevice(DummyList)
    times = []
    with dev:
        start = time()
        while time() - start < 10:
            t0 = default_timer()
            dat = dev.read()
            t1 = default_timer()
            if dat.num2 is not None:
                dff = t1 - t0
                # print(dff)
                # print(dat.num1.data.shape)
                # print(np.diff(dat.num1.time))
                print(dat.num2.time)
                times.append(np.diff(dat[0].time[::2]))
                sleep(0.016)
    plt.plot(np.hstack(times))
    plt.show()



from ctypes import c_double

import numpy as np

import hid
import usb.core
import usb.util
from toon.input.device import BaseDevice, Obs


class Hand(BaseDevice):
    sampling_frequency = 1000

    class Pos(Obs):
        shape = (15,)
        ctype = c_double

    def __init__(self, blocking=True, **kwargs):
        super(Hand, self).__init__(**kwargs)
        self._sqrt2 = np.sqrt(2)
        self._device = None
        self._buffer = np.full(15, np.nan)
        self.blocking = blocking

    def __enter__(self):
        self._device = hid.device()
        # TODO: we may have different endpoints, make more robust?
        self._device.open(0x16c0, 0x486)
        self._device.set_nonblocking(not self.blocking)
        return self

    def __exit__(self):
        self._device.close()

    def read(self):
        data = self._device.read(46)
        time = self.clock()
        # timestamp, deviation from period, and 20x16-bit analog channels
        data = struct.unpack('>Lh' + 'H' * 20, bytearray(data))
        data = np.array(data, dtype='d')
        data[2:] /= 65535.0
        data[2:] -= 0.5
        self._buffer[0::3] = (data[2::4] - data[3::4])/self._sqrt2
        self._buffer[1::3] = (data[2::4] + data[3::4])/self._sqrt2
        self._buffer[2::3] = data[4::4] + data[5::4]
        return self.Returns(self.Pos(time, data))


# USB demo (should be phenotypically identical to above)
class UsbHand(BaseDevice):
    sampling_frequency = 1000

    class Pos(Obs):
        shape = (15,)
        ctype = c_double

    def __init__(self, **kwargs):
        super(UsbHand, self).__init__(**kwargs)
        self._sqrt2 = np.sqrt(2)
        self._device = None
        self._buffer = np.full(15, np.nan)

    def __enter__(self):
        dev = usb.core.find(idVendor=0x16c0, idProduct=0x486)
        if dev is None:
            raise ValueError('Device not found.')
        # there was definitely more to claiming the device,
        # but I can't find the ref now
        self.ep_in = dev[0][(0, 0)][0]  # get the proper endpoint
        return self

    def __exit__(self):
        pass  # nothing to do?

    def read(self):
        data = self.ep_in.read(self.ep_in.wMaxPacketSize)
        time = self.clock()
        data = struct.unpack('>Lh' + 'H' * 20, data[:46])
        data = np.array(data, dtype='d')
        data[2:] /= 65535.0
        data[2:] -= 0.5
        self._buffer[0::3] = (data[2::4] - data[3::4])/self._sqrt2
        self._buffer[1::3] = (data[2::4] + data[3::4])/self._sqrt2
        self._buffer[2::3] = data[4::4] + data[5::4]
        return self.Returns(self.Pos(time, data))



from toon.input.device import BaseDevice, Obs
import ctypes
from pynput import mouse


class Mouse(BaseDevice):
    class Pos(Obs):
        shape = (2,)
        ctype = int

    class Clicks(Obs):
        shape = (1,)
        ctype = bool

    class Scroll(Obs):
        shape = (1,)
        ctype = ctypes.c_int

    sampling_frequency = 125

    def __init__(self, **kwargs):
        super(Mouse, self).__init__(**kwargs)

    def __enter__(self):
        self.dev = mouse.Listener(on_move=self.on_move,
                                  on_click=self.on_click,
                                  on_scroll=self.on_scroll)
        self.dev.start()
        self.dev.wait()
        self.x_prev = 0
        self.y_prev = 0
        self.data = []
        return self

    def read(self):
        if not self.data:
            return self.Returns()
        ret = self.data.copy()
        self.data = []
        return ret

    def on_move(self, x, y):
        # relative mouse position
        rets = self.Returns(pos=self.Pos(self.clock(), (x - self.x_prev,
                                                        y - self.y_prev)))
        self.x_prev = x
        self.y_prev = y
        self.data.append(rets)

    def on_click(self, x, y, button, pressed):
        rets = self.Returns(clicks=self.Clicks(self.clock(), pressed))
        self.data.append(rets)

    def on_scroll(self, x, y, dx, dy):
        rets = self.Returns(scroll=self.Scroll(self.clock(), dy))
        self.data.append(rets)

    def __exit__(self, *args):
        self.dev.stop()
        self.dev.join()


if __name__ == '__main__':
    import time
    from toon.input.mpdevice import MpDevice
    dev = MpDevice(Mouse)
    with dev:
        start = time.time()
        while time.time() - start < 10:
            dat = dev.read()
            if dat.any():
                print(dat)
            time.sleep(0.016)  # pretend to have a screen



from math import pow, sin, pi
# from https://github.com/mosra/magnum/blob/master/src/Magnum/Animation/Easing.h

pihalf = pi/2.0


def linear(x):
    return x


def step(x):
    return 0.0 if x < 0.5 else 1.0


def smoothstep(x):
    if x <= 0.0:
        return 0.0
    if x >= 1.0:
        return 1.0
    return (3.0 - 2.0 * x) * pow(x, 2)


def smootherstep(x):
    if x <= 0.0:
        return 0.0
    if x >= 1.0:
        return 1.0
    return pow(x, 3) * (x * (x * 6.0 - 15.0) + 10.0)


def quadratic_in(x):
    return pow(x, 2)


def quadratic_out(x):
    return -x * (x - 2.0)


def quadratic_in_out(x):
    if x < 0.5:
        return 2.0*pow(x, 2)
    return 1.0 - 2.0*pow(1.0 - x, 2)


def exponential_in(x):
    return 0.0 if x <= 0.0 else pow(2.0, 10.0 * (x - 1.0))


def exponential_out(x):
    return 1.0 if x >= 1.0 else 1.0 - pow(2.0, -10.0 * x)


def exponential_in_out(x):
    if x <= 0.0:
        return 0.0
    if x >= 1.0:
        return 1.0
    if x < 0.5:
        return 0.5 * pow(2.0, 20.0*x - 10.0)
    return 1.0 - 0.5 * pow(2.0, 10.0 - 20.0 * x)


def elastic_in(x):
    return pow(2.0, 10.0 * (x - 1.0)) * sin(13.0 * pihalf * x)


def elastic_out(x):
    return 1.0 - pow(2.0, -10.0 * x) * sin(13.0 * pihalf * (x + 1.0))


def elastic_in_out(x):
    if x < 0.5:
        return 0.5 * pow(2.0, 10.0 * (2.0 * x - 1.0)) * sin(13.0 * pi * x)
    return 1.0 - 0.5 * pow(2.0, 10.0 * (1.0 - 2.0 * x)) * sin(13.0 * pi * x)


def back_in(x):
    return x * (x * x - sin(pi * x))


def back_out(x):
    inv = 1.0 - x
    return 1 - inv*(inv*inv - sin(pi*inv))


def back_in_out(x):
    if x < 0.5:
        x2 = 2.0 * x
        return 0.5 * x2 * (x2 * x2 - sin(pi * x2))
    inv = 2.0 - 2.0 * x
    return 1.0 - 0.5 * inv * (inv * inv - sin(pi * inv))


def bounce_out(x):
    if x == 0:
        return 0
    if x < 4.0/11.0:
        return 121.0*x*x/16.0
    if x < 8.0/11.0:
        return 363.0/40.0*x*x - 99.0/10.0 * x + 17.0/5.0
    if x < 9.0/10.0:
        return 4356.0/361.0*x*x - 35442.0/1805.0*x + 16061.0/1805.0
    if x == 1:
        return 1
    return 54.0/5.0 * x*x - 513.0/25.0*x + 268.0/25.0


def bounce_in(x):
    return 1.0 - bounce_out(1.0 - x)


def bounce_in_out(x):
    if x < 0.5:
        return 0.5 * bounce_in(2*x)
    return 0.5 * bounce_out(2.0 * x - 1) + 0.5



from time import sleep

import numpy as np
from pytest import approx, raises
from tests.input.mockdevices import Dummy, DummyList, SingleResp, Timebomb

from toon.input.clock import mono_clock
from toon.input.mpdevice import MpDevice

# bump up the sampling frequency for tests
Dummy.sampling_frequency = 1000
DummyList.sampling_frequency = 1000


def test_device_single():
    # single device with two data sources
    dev = MpDevice(Dummy)
    dev.start()
    sleep(0.2)
    res = dev.read()
    dev.stop()
    assert(len(res.num1.time) > 5)
    assert(len(res.num1.time) == res.num1.shape[0])
    assert(res.num1.shape[1] == 5)
    assert(res.num2.dtype == np.int32)


def test_single_resp():
    dev = MpDevice(SingleResp)
    dev.start()
    sleep(0.2)
    res = dev.read()
    dev.stop()
    assert(isinstance(res, np.ndarray))
    assert(type(res.time) is np.ndarray)


def test_ringbuffer():
    original_fs = SingleResp.sampling_frequency
    SingleResp.sampling_frequency = 10
    dev = MpDevice(SingleResp)
    with dev:
        sleep(5)
        res = dev.read()
    assert(all(np.diff(res) == 1))
    assert(all(np.diff(res.time) > 0))
    SingleResp.sampling_frequency = original_fs


def test_have_all_data():
    dev = MpDevice(SingleResp)
    datae = []
    times = []
    with dev:
        for i in range(100):
            sleep(0.2)
            data = dev.read()
            if data is not None:
                datae.append(np.copy(data))
                times.append(np.copy(data.time))
    times = np.hstack(times)
    datae = np.hstack(datae)
    # check time is always increasing
    assert((np.diff(times) > 0).all())
    # check data (should be monotonically increasing)
    assert((np.diff(datae) == 1).all())


def test_device_list():
    # two observations per read on the device
    dev = MpDevice(DummyList)
    dev.start()
    sleep(0.2)
    res = dev.read()
    dev.stop()
    assert(len(res.num1.time) > 5)
    assert(len(res.num1.time) == res.num1.shape[0])
    assert(res.num1.shape[1] == 5)
    assert(res.num2.dtype == np.int32)
    assert(res.num1.shape[0] > res.num2.shape[1])


def test_restart():
    # start & stop device
    dev = MpDevice(Dummy)
    dev.start()
    sleep(0.2)
    res = dev.read()
    dev.stop()
    dev.start()
    sleep(0.2)
    res2 = dev.read()
    dev.stop()
    assert(res.any() and res2.any())


def test_context():
    # device as context manager
    dev = MpDevice(Dummy)
    with dev:
        sleep(0.2)
        res = dev.read()
    assert(len(res.num1.time) > 5)
    assert(len(res.num1.time) == res.num1.shape[0])
    assert(res.num1.shape[1] == 5)
    assert(res.num2.dtype == np.int32)


def test_multi_devs():
    # 2+ devices at once (each gets own process)
    dev1 = MpDevice(Dummy)
    dev2 = MpDevice(Dummy)
    with dev1, dev2:
        sleep(0.1)
        res1 = dev1.read()
        res2 = dev2.read()

    assert(res1.num1.time is not None)
    assert(res2.num1.time is not None)


def test_remote_clock():
    # does the clock origin match?
    # the most recent reading should've been within
    # one sampling period
    dev = MpDevice(SingleResp, clock=mono_clock.getTime)
    sleep(0.5)
    with dev:
        sleep(0.1)
        res = dev.read()
        time = mono_clock.get_time()
    assert(time - res.time[-1] == approx(1.0/SingleResp.sampling_frequency, abs=1e-3))


def test_already_closed():
    dev = MpDevice(SingleResp)
    dev.start()
    res = dev.read()
    dev.stop()
    with raises(ValueError):
        dev.read()


def test_catch_remote_err():
    dev = MpDevice(Timebomb)
    with dev:
        sleep(0.1)
        with raises(ValueError):
            res = dev.read()



from toon.input.device import BaseDevice, Obs
import ctypes
from pynput import keyboard


class Keyboard(BaseDevice):
    class Key(Obs):
        shape = (1,)
        ctype = ctypes.c_char

    class Press(Obs):
        shape = (1,)
        ctype = ctypes.c_bool

    sampling_frequency = 100

    def __init__(self, keys=None, **kwargs):
        if keys is None:
            raise ValueError('Specify keys.')
        self.keys = keys
        super(Keyboard, self).__init__(**kwargs)

    def __enter__(self):
        self.dev = keyboard.Listener(on_press=self.on_press,
                                     on_release=self.on_release)
        self.data = []
        self._on = []
        self.dev.start()
        self.dev.wait()
        return self

    def read(self):
        if not self.data:
            return self.Returns()
        ret = self.data.copy()
        self.data = []
        return ret

    def on_press(self, key):
        time = self.clock()
        if not isinstance(key, keyboard.Key):
            # check if key is of interest, and
            # ignore repeating keys
            if key.char in self.keys and key.char not in self._on:
                rets = self.Returns(key=self.Key(time, key.char),
                                    press=self.Press(time, True))
                self.data.append(rets)
                self._on.append(key.char)

    def on_release(self, key):
        time = self.clock()
        if not isinstance(key, keyboard.Key):
            if key.char in self.keys and key.char in self._on:
                rets = self.Returns(key=self.Key(time, key.char),
                                    press=self.Press(time, False))
                self.data.append(rets)
                self._on.remove(key.char)

    def __exit__(self, *args):
        self.dev.stop()
        self.dev.join()


if __name__ == '__main__':
    import time
    from toon.input.mpdevice import MpDevice
    dev = MpDevice(Keyboard, keys=['a', 's', 'd', 'f'])
    with dev:
        start = time.time()
        while time.time() - start < 10:
            dat = dev.read()
            if dat.any():
                print(dat)
            time.sleep(0.016)
